import platform
import subprocess
import sys
import tarfile
import urllib.request
import io
import os
from urllib.error import HTTPError
from os import path
from subprocess import Popen

from importlib_metadata import version
from pathlib import Path

app_version = version("blindai")


class MockServer:
    """Mock server process handle"""

    def __init__(self, process, _private=False):
        if not _private:
            raise NotImplementedError()
        self._process = process

    def stop(self) -> bool:
        """Stop BlindAI mock server.

        This method will kill the running server, if the provided MockServer is still running.
        Args:
            self (MockServer): Mock server process handle
        Return:
            bool, True if the server was successfully stopped.
        Raises:
            None
        """

        if self._process is not None and self._process.poll() is None:
            print("Stopping BlindAI mock server...")
            self._process.kill()
            return True
        else:
            print("BlindAI mock server already stopped")
            return False


class NotFoundError(Exception):
    """This exception is raised when there was an error opening an URL.
    Args:
        Args:
        message (str): Error message.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def _extract_tar(data):
    with tarfile.open(fileobj=io.BytesIO(data), mode="r:gz") as tar:
        tar.extractall(Path.cwd() / "bin")


def _handle_download(path: Path, url: str, name: str, error_msg: str):
    if not path.exists():
        print(f"Downloading {name}...")
        try:
            response = urllib.request.urlopen(url)
            _extract_tar(response.read())
        except HTTPError as e:
            raise NotFoundError(
                "{}. Exact error code: {}".format(error_msg, e.code)
            ) from None
    else:
        print("{} already installed".format(name))


def _start_server(blindai_path: Path) -> MockServer:
    blindai_path.chmod(0o755)
    process = subprocess.Popen([blindai_path], env=os.environ)
    return MockServer(process, _private=True)


def start_mock_server() -> MockServer:
    """Start BlindAI mock server for testing.
    The method will download BlindAI Preview's mock server binary if needed.
    The mock server is then started allowing to run the rest of your Google Colab/Jupyter Notebook environment.
    Args:
        None
    Return:
        MockServer object, the process of the running server.
    Raises:
        NotFoundError: Will be raised if one of the URL the wheel will try to access is invalid. This might mean that either there is no available binary of BlindAI's server.
        Other exceptions might be raised by zipfile or urllib.request.
    """
    blindai_path = Path.cwd() / "bin" / "blindai_mock_server"

    arch = platform.machine()
    if arch == "AMD64":
        arch = "x86_64"

    if not (arch == "x86_64" and sys.platform == "linux"):
        raise RuntimeError(f"Unsupported system : {platform.machine()}-{sys.platform}")

    blindai_url = f"https://github.com/mithril-security/blindai/releases/download/v{app_version}/blindai_mock_server-{app_version}-x86_64-unknown-linux-gnu.tgz"

    _handle_download(
        blindai_path,
        blindai_url,
        f"BlindAI mock server (version {app_version})",
        "The release might not be available yet for the current version",
    )
    process = _start_server(blindai_path)
    return process



# Copyright 2022 Mithril Security. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pathlib
from ._dcap_attestation import validate_attestation, AttestationError, Collateral
from .utils import *

from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple, Union

import os
import contextlib
import socket
import sys

import numpy as np
import cbor2 as cbor

from hashlib import sha256
import platform
import getpass
import logging
import tempfile
import requests
from requests.adapters import HTTPAdapter
from importlib_metadata import version
import warnings

app_version = version("blindai")

CONNECTION_TIMEOUT = 10


class SimulationModeWarning(Warning):
    pass


class ModelDatumType(IntEnum):
    F32 = 0
    F64 = 1
    I32 = 2
    I64 = 3
    U32 = 4
    U64 = 5
    U8 = 6
    U16 = 7
    I8 = 8
    I16 = 9
    Bool = 10


format_per_item = {
    ModelDatumType.F32: "<f4",
    ModelDatumType.F64: "<f8",
    ModelDatumType.I32: "<i4",
    ModelDatumType.I64: "<i8",
    ModelDatumType.U32: "<u4",
    ModelDatumType.U64: "<u8",
    ModelDatumType.U8: "<u1",
    ModelDatumType.U16: "<u2",
    ModelDatumType.I8: "<i1",
    ModelDatumType.I16: "<i2",
    ModelDatumType.Bool: "?",
}


def serialize_tensor(tensor: np.ndarray, type: ModelDatumType) -> bytes:
    return np.array(tensor).astype(format_per_item[type], casting="equiv").tobytes()


def deserialize_tensor(data: bytes, type: ModelDatumType) -> np.ndarray:
    return np.frombuffer(data, dtype=format_per_item[type])


class TensorInfo:
    fact: List[int]
    datum_type: ModelDatumType
    node_name: str

    def __init__(self, fact, datum_type, node_name=None):
        self.fact = fact
        self.datum_type = (
            ModelDatumType[datum_type]
            if isinstance(datum_type, str)
            else ModelDatumType(datum_type)
        )
        self.node_name = node_name


class Tensor:
    """Tensor class to convert serialized tensors into convenients objects."""

    info: Union[TensorInfo, dict]
    bytes_data: bytes

    def __init__(self, info: Union[TensorInfo, dict], bytes_data: bytes):
        self.info = info
        self.bytes_data = bytes_data

    def as_flat(self) -> list:
        """Convert the prediction calculated by the server to a flat python
        list."""
        return self.as_numpy().tolist()

    def as_numpy(self):
        """Convert the prediction calculated by the server to a numpy array."""

        arr = deserialize_tensor(self.bytes_data, self.info.datum_type)
        arr.shape = self.shape
        return arr

    def as_torch(self):
        """Convert the prediction calculated by the server to a Torch Tensor.

        As torch is heavy it's an optional dependency of the project, and is
        imported only when needed.

        Raises: ImportError if torch isn't installed
        """
        try:
            import torch
        except ImportError as e:
            raise ImportError(
                "torch not installed, please install with pip install blindai[torch]"
            ) from e
        arr = torch.tensor(self.as_numpy())
        return arr.view(self.shape)

    @property
    def shape(self) -> tuple:
        if isinstance(self.info, TensorInfo):
            return tuple(self.info.fact)
        return self.info["fact"]

    @property
    def datum_type(self) -> ModelDatumType:
        if isinstance(self.info, TensorInfo):
            return self.info.datum_type
        return self.info["datum_type"]


@dataclass
class UploadModel:
    model: List[int]
    length: int
    model_name: str
    optimize: bool
    client_info: "_ClientInfo"

    def __init__(
        self,
        model,
        length,
        client_info,
        model_name="",
        optimize=True,
    ):
        self.model = model
        self.length = length
        self.model_name = model_name
        self.optimize = optimize
        self.client_info = client_info


@dataclass
class RunModel:
    model_id: str
    model_hash: str
    inputs: List[Tensor]
    client_info: Optional["_ClientInfo"]

    def __init__(self, model_id, model_hash, inputs, client_info=None):
        self.model_id = model_id
        self.model_hash = model_hash
        self.inputs = inputs
        self.client_info = client_info


@dataclass
class DeleteModel:
    model_id: str

    def __init__(self, model_id):
        self.model_id = model_id


@dataclass
class SendModelReply:
    hash: bytes
    model_id: str

    def __init__(self, **entries):
        self.__dict__.update(entries)


@dataclass
class RunModelReply:
    outputs: List[Any]

    def __init__(self, **entries):
        self.__dict__.update(entries)


@dataclass
class UploadResponse:
    model_id: str
    hash: bytes


@dataclass
class RunModelResponse:
    output: List[Tensor]


@dataclass
class _ClientInfo:
    uid: str
    platform_name: str
    platform_arch: str
    platform_version: str
    platform_release: str
    user_agent: str
    user_agent_version: str
    is_colab: bool

    def __init__(
        self,
        uid,
        platform_name,
        platform_arch,
        platform_version,
        platform_release,
        user_agent,
        user_agent_version,
        is_colab,
    ):
        self.uid = uid
        self.platform_name = platform_name
        self.platform_arch = platform_arch
        self.platform_version = platform_version
        self.platform_release = platform_release
        self.user_agent = user_agent
        self.user_agent_version = user_agent_version
        self.is_colab = is_colab

    def __iter__(self) -> dict:
        return {
            "uid": self.uid,
            "platform_name": self.platform_name,
            "platform_arch": self.platform_arch,
            "platform_version": self.platform_version,
            "platform_release": self.platform_release,
            "user_agent": self.user_agent,
            "user_agent_version": self.user_agent_version,
            "is_colab": self.is_colab,
        }


def dtype_to_numpy(dtype: ModelDatumType) -> str:
    """Convert a ModelDatumType to a numpy type.

    Raises:
        ValueError: if numpy doesn't support dtype
    """
    translation_map = {
        ModelDatumType.F32: "float32",
        ModelDatumType.F64: "float64",
        ModelDatumType.I32: "int32",
        ModelDatumType.I64: "int64",
        ModelDatumType.U32: "uint32",
        ModelDatumType.U64: "uint64",
        ModelDatumType.U8: "uint8",
        ModelDatumType.U16: "uint16",
        ModelDatumType.I8: "int8",
        ModelDatumType.I16: "int16",
        ModelDatumType.Bool: "bool",
    }
    if dtype not in translation_map:
        raise ValueError(f"Numpy does not support datum type {dtype}.")
    return translation_map[dtype]


def dtype_to_torch(dtype: ModelDatumType) -> str:
    """Convert a ModelDatumType to a torch type.

    Raises:
        ValueError: if torch doesn't support dtype
    """
    # Torch does not support unsigned ints except u8.
    translation_map = {
        ModelDatumType.F32: "float32",
        ModelDatumType.F64: "float64",
        ModelDatumType.I32: "int32",
        ModelDatumType.I64: "int64",
        ModelDatumType.U8: "uint8",
        ModelDatumType.I8: "int8",
        ModelDatumType.I16: "int16",
        ModelDatumType.Bool: "bool",
    }
    if dtype not in translation_map:
        raise ValueError(f"Torch does not support datum type {dtype}.")
    return translation_map[dtype]


def translate_dtype(dtype: Any) -> ModelDatumType:
    """
    Convert torch, numpy or litteral types to ModelDatumType
    Raises:
        ValueError: if dtype is erroneous or not supported
    """

    if isinstance(dtype, ModelDatumType):
        return dtype

    elif type(dtype).__module__ == "numpy" and type(dtype).__name__.startswith("dtype"):
        numpy_dtype_translation = {
            "float32": ModelDatumType.F32,
            "float64": ModelDatumType.F64,
            "int32": ModelDatumType.I32,
            "int64": ModelDatumType.I64,
            "uint32": ModelDatumType.U32,
            "uint64": ModelDatumType.U64,
            "uint8": ModelDatumType.U8,
            "uint16": ModelDatumType.U16,
            "int8": ModelDatumType.I8,
            "int16": ModelDatumType.I16,
            "bool": ModelDatumType.Bool,
        }
        if str(dtype) not in numpy_dtype_translation:
            raise ValueError(f"Numpy dtype {str(dtype)} is not supported.")
        return numpy_dtype_translation[str(dtype)]

    if type(dtype).__module__ == "torch" and type(dtype).__name__ == "dtype":
        # Torch does not support unsigned ints except u8.
        torch_dtype_translation = {
            "torch.float32": ModelDatumType.F32,
            "torch.float64": ModelDatumType.F64,
            "torch.int32": ModelDatumType.I32,
            "torch.int64": ModelDatumType.I64,
            "torch.uint8": ModelDatumType.U8,
            "torch.int8": ModelDatumType.I8,
            "torch.int16": ModelDatumType.I16,
            "torch.bool": ModelDatumType.Bool,
        }
        if str(dtype) not in torch_dtype_translation:
            raise ValueError(f"Torch dtype {str(dtype)} is not supported.")
        return torch_dtype_translation[str(dtype)]

    if isinstance(dtype, str):
        str_dtype_translation = {
            "float32": ModelDatumType.F32,
            "f32": ModelDatumType.F32,
            "float64": ModelDatumType.F64,
            "f64": ModelDatumType.F64,
            "int32": ModelDatumType.I32,
            "i32": ModelDatumType.I32,
            "int64": ModelDatumType.I64,
            "i64": ModelDatumType.I64,
            "uint32": ModelDatumType.U32,
            "u32": ModelDatumType.U32,
            "uint64": ModelDatumType.U64,
            "u64": ModelDatumType.U64,
            "uint8": ModelDatumType.U8,
            "u8": ModelDatumType.U8,
            "uint16": ModelDatumType.U16,
            "u16": ModelDatumType.U16,
            "int8": ModelDatumType.I8,
            "i8": ModelDatumType.I8,
            "int16": ModelDatumType.I16,
            "i16": ModelDatumType.I16,
            "bool": ModelDatumType.Bool,
        }
        if dtype.lower() not in str_dtype_translation:
            raise ValueError(f"Datum type {dtype} is not understood.")
        return str_dtype_translation[dtype.lower()]

    raise ValueError(
        f"DatumType instance {type(dtype).__module__}.{type(dtype).__name__} not supported"
    )


def _is_torch_tensor(tensor) -> bool:
    return type(tensor).__module__ == "torch" and type(tensor).__name__ == "Tensor"


def _is_numpy_array(tensor) -> bool:
    return type(tensor).__module__ == "numpy" and type(tensor).__name__ == "ndarray"


def translate_tensor(
    tensor: Any, or_dtype: ModelDatumType, or_shape: Tuple, name=None
) -> Tensor:
    """Put the flat/numpy/torch tensor into a Tensor object.

    Args:
        tensor: flat/numpy/torch tensor
        or_dtype: ignored if tensor isn't flat. dtype of the tensor.
        or_shape: ignored if tensor isn't flat. shape of the tensor.
    Raises:
        ValueError: if tensor format is not one of flat/numpy/torch
        ValueError: if tensor's dtype is not supported
    Returns:
        Tensor: the serialized tensor
    """
    if _is_torch_tensor(tensor):
        info = TensorInfo(tensor.shape, translate_dtype(tensor.dtype), name)
        iterable = tensor.flatten().numpy()

    elif _is_numpy_array(tensor):
        info = TensorInfo(tensor.shape, translate_dtype(tensor.dtype), name)
        iterable = tensor.flatten()

    else:
        # Input is flatten tensor.
        if not isinstance(tensor, list):
            raise ValueError(
                f"Input tensor has an unsupported type: {type(tensor).__module__}.{type(tensor).__name__}"
            )

        info = TensorInfo(or_shape, translate_dtype(or_dtype), name)
        iterable = np.array(tensor, dtype=dtype_to_numpy(or_dtype))

    if or_dtype is not None and or_dtype != info.datum_type:
        raise ValueError(
            f"Given tensor has dtype {str(tensor.dtype)}, but {or_dtype} was expected."
        )

    # todo validate tensor content, dtype and shape
    return Tensor(info.__dict__, serialize_tensor(iterable, info.datum_type))


def translate_tensors(tensors, dtypes, shapes) -> List[dict]:
    """Put the flat/numpy/torch tensors into a list of Tensor objects.

    Args:
        tensor: list or dict of flat/numpy/torch tensors
        dtypes: ignored if tensors aren't flat. list or dict of dtypes of the tensors.
        or_shape: ignored if tensor aren't flat. list or dict of shapes of the tensors.
    Returns:
        List[dict]: the serialized tensors as a list of dicts.
    """
    serialized_tensors = []

    # dict of tensors is the safe mean of passing inputs
    # if it's a dict of flat tensors, dtypes and shapes must be dicts as well
    #
    # list of {numpy/torch/flat} tensors are valid inputs, and are treated as multiple inputs
    # direct numpy/torch/flat tensors are valid inputs, and are treated as a single input, which
    #  will be wrapped into a 1-el list on the folowing statement
    #
    # flat list means list[int], and is the flattened tensor
    #  this means that you must specify dtype/shape for this tensor! on the other cases, it's redundant
    # (todo: accept iterables instead of flat list only)
    #
    # mental note
    # - anything not list should be wrapped into [X]
    # - list[int] should be wrapped into [X]
    # - but! list[list[int]] is should be unchanged

    if isinstance(tensors, dict):
        for name, tensor in tensors.items():
            or_dtype = dtypes[name] if dtypes is not None else None
            or_shape = shapes[name] if shapes is not None else None
            serialized_tensors.append(
                translate_tensor(tensor, or_dtype, or_shape, name).__dict__
            )
    else:
        # if arg is not a list of (list/numpy.array/torch.tensor), wrap it into a list
        if not isinstance(tensors, list) or (
            len(tensors) > 0
            and not (
                isinstance(tensors[0], list)
                or _is_torch_tensor(tensors[0])
                or _is_numpy_array(tensors[0])
            )
        ):
            tensors = [tensors]
        if dtypes is not None and not isinstance(dtypes, list):
            dtypes = [dtypes]
        if shapes is not None and not isinstance(shapes, list):
            shapes = [shapes]

        for i, tensor in enumerate(tensors):
            or_dtype = dtypes[i] if dtypes is not None and len(dtypes) > i else None
            or_shape = shapes[i] if shapes is not None and len(shapes) > i else None
            serialized_tensors.append(
                translate_tensor(tensor, or_dtype, or_shape).__dict__
            )

    return serialized_tensors


class BlindAiConnection(contextlib.AbstractContextManager):
    """A class to represent a connection to a BlindAi server."""

    _conn: requests.Session

    def __init__(
        self,
        addr: str,
        unattested_server_port: int,
        attested_server_port: int,
        model_management_port: int,
        hazmat_manifest_path: Optional[pathlib.Path],
        hazmat_http_on_unattested_port: bool,
        simulation_mode: bool,
        use_cloud_manifest: bool,
    ):
        """Connect to a BlindAi service.

        Please refer to the connect function for documentation.

        Args:
            addr (str):
            unattested_server_port (int):
            attested_server_port (int):
            model_management_port (int):
            hazmat_manifest_path (Optional[pathlib.Path]):
            hazmat_http_on_unattested_port (bool):
            simulation_mode (bool):
        Returns:
        """

        if simulation_mode:
            warnings.warn(
                (
                    "BlindAI is running in simulation mode. "
                    "This mode is provided solely for testing purposes. "
                    "It does not provide any security since there is no SGX enclave. "
                    "The simulation mode MUST NOT be used in production."
                ),
                SimulationModeWarning,
            )

        uname = platform.uname()

        self.client_info = _ClientInfo(
            uid=sha256((socket.gethostname() + "-" + getpass.getuser()).encode("utf-8"))
            .digest()
            .hex(),
            platform_name=uname.system,
            platform_arch=uname.machine,
            platform_version=uname.version,
            platform_release=uname.release,
            user_agent="blindai_python",
            user_agent_version=app_version,
            is_colab="google.colab" in sys.modules,
        )

        if hazmat_http_on_unattested_port:
            self._unattested_url = f"http://{addr}:{unattested_server_port}"
        else:
            self._unattested_url = f"https://{addr}:{unattested_server_port}"

        self._attested_url = f"https://{addr}:{attested_server_port}"

        self._model_management_url = f"https://{addr}:{model_management_port}"

        # This adapter makes it possible to connect
        # to the server via a different hostname
        # that the one included in the certificate i.e. blindai-srv
        # For instance we can use it to connect to the server via the
        # domain / IP provided to connect(). See below
        class CustomHostNameCheckingAdapter(HTTPAdapter):
            def cert_verify(self, conn, url, verify, cert):
                conn.assert_hostname = "blindai-srv"
                return super(CustomHostNameCheckingAdapter, self).cert_verify(
                    conn, url, verify, cert
                )

        s = requests.Session()
        # Always raise an exception when HTTP returns an error code for the unattested connection
        # Note : we might want to do the same for the attested connection ?
        s.hooks = {"response": lambda r, *args, **kwargs: r.raise_for_status()}
        req = s.get(self._unattested_url)
        cert = cbor.loads(req.content)
        if not simulation_mode and "mock" in req.headers["Server"]:
            raise AttestationError(
                "The BlindAI server is a mock. You can only connect to it in simulation mode."
            )

        if not simulation_mode:
            try:
                quote = cbor.loads(s.get(f"{self._unattested_url}/quote").content)
                collateral = cbor.loads(
                    s.get(f"{self._unattested_url}/collateral").content
                )
                try:
                    collateral = Collateral(**collateral)
                except TypeError as e:
                    raise AttestationError("Bad attestation collateral from the server")

                validate_attestation(
                    quote,
                    collateral,
                    cert,
                    manifest_path=hazmat_manifest_path,
                    use_cloud_manifest=use_cloud_manifest,
                )
            except AttestationError as e:
                raise
            except Exception as e:
                raise AttestationError("Attestation verification failed")

        # requests (http library) takes a path to a file containing the CA
        # there is no easy way to give the CA as a string/bytes directly
        # therefore a temporary file with the certificate content
        # has to be created.

        attested_server_cert_file = tempfile.NamedTemporaryFile(mode="wb")
        attested_server_cert_file.write(cert_der_to_pem(cert))
        attested_server_cert_file.flush()
        # the file should not be close until the end of BlindAiConnection
        # so we store it in the object (else it might get garbage collected)
        self.attested_cert_file = attested_server_cert_file

        attested_conn = requests.Session()
        attested_conn.verify = attested_server_cert_file.name
        attested_conn.mount(self._attested_url, CustomHostNameCheckingAdapter())
        attested_conn.mount(self._model_management_url, CustomHostNameCheckingAdapter())

        # finally try to connect to the enclave
        try:
            attested_conn.get(self._attested_url)
        except Exception as e:
            raise AttestationError("Cannot establish secure connection to the enclave")

        self._conn = attested_conn

    def upload_model(
        self,
        model: str,
        model_name: Optional[str] = None,
        optimize: bool = True,
    ) -> UploadResponse:
        """Upload an inference model to the server.

        The provided model needs to be in the Onnx format.

        ***Security & confidentiality warnings:***
            model: The model sent on a Onnx format is encrypted in transit via TLS (as all connections).
            It may be subject to inference Attacks if an adversary is able to query the trained model
            repeatedly to determine whether or not a particular example is part of the trained dataset model.
        Args:
            model (str): Path to Onnx model file.
            model_name (Optional[str], optional): Name of the model.
                Used for you to identify the model, but won't be used by the server (a random UUID will be assigned to your model for the inferences).
            optimize (bool): Whether tract (our inference engine) should optimize the model or not.
                Optimzing should only be turned off when you are encountering issues loading your model.
        Raises:
            HttpError: raised by the requests lib to relay server side errors
            ValueError: raised when inputs sanity checks fail
        Returns:
            UploadResponse: The response object.
        """
        if model_name is None:
            model_name = os.path.basename(model)

        with open(model, "rb") as f:
            model_bytes = f.read()

        length = len(model_bytes)

        data = UploadModel(
            model=list(model_bytes),
            length=length,
            model_name=model_name,
            optimize=optimize,
            client_info=self.client_info.__dict__,
        )
        bytes_data = cbor.dumps(data.__dict__)
        r = self._conn.post(f"{self._model_management_url}/upload", data=bytes_data)
        r.raise_for_status()
        send_model_reply = SendModelReply(**cbor.loads(r.content))
        ret = UploadResponse(
            model_id=send_model_reply.model_id, hash=send_model_reply.hash
        )
        return ret

    def run_model(
        self,
        model_id: str = "",
        model_hash: str = "",
        input_tensors: Optional[Union[List, Dict]] = None,
        dtypes: Optional[List[ModelDatumType]] = None,
        shapes: Optional[Union[List[List[int]], List[int]]] = None,
    ) -> RunModelResponse:
        """Send data to the server to make a secure inference.

        The data provided must be in a list, as the tensor will be rebuilt inside the
        server.

        ***Security & confidentiality warnings:***
            model_id: hash of the Onnx model uploaded. the given hash is return via gRPC through the proto files.
            It's a SHA-256 hash that is generated each time a model is uploaded.
            tensors: protected in transit and protected when running it on the secure enclave.
            In the case of a compromised OS, the data is isolated and confidential by SGX design.

        Args:
            model_id (str): If set, will run a specific model.
            model_hash (str): hash of the Onnx model uploaded. If no uuid was provided, the server will try to find a model matching this hash
                input_tensors (Union[List[Any], List[List[Any]]))): The input data. It must be an array of numpy,
                tensors or flat list of the same type datum_type specified in `upload_model`.
            dtypes (Union[List[ModelDatumType], ModelDatumType], optional): The type of data
                of the data you want to upload. Only required if you are uploading flat lists, will be ignored
                if you are uploading numpy or tensors (this info will be extracted directly from the tensors/numpys).
            shapes (Union[List[List[int]], List[int]], optional): The shape of the data you want to upload.
                Only required if you are uploading flat lists, will be ignored if you are uploading numpy
                or tensors (this info will be extracted directly from the tensors/numpys).
        Raises:
            HttpError: raised by the requests lib to relay server side errors
            ValueError: raised when inputs sanity checks fail
        Returns:
            RunModelResponse: The response object.
        """
        # Run Model Request and Response

        if not model_id and not model_hash:
            raise ValueError("You must provide at least one model_id or model_hash")
        if model_id and model_hash:
            raise ValueError(
                "You cannot provide a model_id and a model_hash in the same time"
            )

        tensors = translate_tensors(input_tensors, dtypes, shapes)
        run_data = RunModel(
            model_hash=model_hash,
            model_id=model_id,
            inputs=tensors,
            client_info=self.client_info.__dict__,
        )
        bytes_run_data = cbor.dumps(run_data.__dict__)
        r = self._conn.post(f"{self._attested_url}/run", data=bytes_run_data)
        r.raise_for_status()
        run_model_reply = RunModelReply(**cbor.loads(r.content))

        ret = RunModelResponse(
            output=[
                Tensor(TensorInfo(**output["info"]), output["bytes_data"])
                for output in run_model_reply.outputs
            ]
        )
        return ret

    def delete_model(self, model_id: str):
        """Delete a model in the inference server.

        This may be used to free up some memory. If you did not specify that you
        wanted your model to be saved on the server, please note that the model will
        only be present in memory, and will disappear when the server close.

        **Security & confidentiality warnings: **
            model_id: The deletion of a model does only relies on the `model_id`.
            It doesn't relies on a session token or anything, hence if the `model_id` is known,
            it's deletion is possible.

        Args:
            model_id (str): The id of the model to remove.
        Raises:
            HttpError: raised by the requests lib to relay server side errors
            ValueError: raised when inputs sanity checks fail
        """
        delete_data = DeleteModel(model_id=model_id)
        bytes_delete_data = cbor.dumps(delete_data.__dict__)
        r = self._conn.post(f"{self._model_management_url}/delete", bytes_delete_data)
        r.raise_for_status()

    def close(self):
        self._conn.close()

    def __enter__(self):
        """Return the BlindAiConnection upon entering the runtime context."""
        return self

    def __exit__(self, *args):
        """Close the connection to BlindAI server."""
        self.close()


from functools import wraps


def connect(
    addr: str,
    unattested_server_port: int = 9923,
    attested_server_port: int = 9924,
    model_management_port: int = 9925,
    hazmat_manifest_path: Optional[pathlib.Path] = None,
    hazmat_http_on_unattested_port=False,
    simulation_mode: bool = False,
    use_cloud_manifest: bool = False,
) -> BlindAiConnection:
    """Connect to a BlindAi server.

    Args:
        addr (str): The address of BlindAI server you want to connect to.
            It can be a domain (such as "example.com" or "localhost") or an IP
        unattested_server_port (int, optional): The unattested server port number. Defaults to 9923.
        attested_server_port (int, optional): The attested server port number. Defaults to 9924.
        model_management_port (int, optional): The model management port. Needs to be specified if the server only accepts model upload/deletion locally. Defaults to 9924.
        hazmat_manifest_path (Optional[pathlib.Path], optional):  Path to the Manifest.toml which describes
            which enclave are to be accepted.
            Defaults to the built-in Manifest.toml provided by Mithril Security as part of the Python package.
            You can override the default by providing a path to your own Manifest.toml
            Caution: Changing the manifest can impact the security of the solution.
        hazmat_http_on_unattested_port (bool, optional): If set to True, the client will request the attestation elements of
            the server using a plain HTTP connection instead of a more secure HTTPS connection. Defaults to False.
            Caution: This parameter should never be set to True in production. Using a HTTPS connection is critical to
            get a graceful degradation in case of a failure of the Intel SGX attestation.
        simulation_mode (bool, optional): If set to True, BlindAI will work in simulation mode.
            Caution: In simulation, BlindAI does not provide any security since there is no SGX enclave.
            This mode SHOULD NEVER be enabled in production.
            Defaults to False (production mode)
        use_cloud_manifest (bool, optional): If set to True, the manifest for the local model management version (aka the cloud version) will be used.

     Raises:
        requests.exceptions.RequestException: If a network or server error occurs
        ValueError: raised when inputs sanity checks fail
        IdentityError: raised when the enclave signature does not match the enclave signature expected in the manifest
        EnclaveHeldDataError: raised when the expected enclave held data does not match the one in the quote
        QuoteValidationError: raised when the returned quote is invalid (TCB outdated, not signed by the hardware provider...).
        AttestationError: raised when the attestation is not valid (enclave settings mismatching, debug mode unallowed...)

    Returns:
        BlindAiConnection: An object representing an active connection to a BlindAi server
    """

    return BlindAiConnection(
        addr,
        unattested_server_port,
        attested_server_port,
        model_management_port,
        hazmat_manifest_path,
        hazmat_http_on_unattested_port,
        simulation_mode,
        use_cloud_manifest,
    )



import os
import wave
import ffmpeg
from typing import Union
import subprocess
import multiprocessing
import numpy as np
import cbor2 as cbor
import tempfile

"""
This is how I would normally do it:
1. We can use a different protocol say `gRPC` or `HTTP` to stream the wave file.

    Using this structure:
        {
            "nframes": int,
            "sampwidth": int,
            "framerate": int,
            "nchannels": int,
            "comptype": str,
            "compname": str,
            "frames": bytes,
        }

2. Then, we can use multiprocessing to read the file and send it to the other process

3. The other process can then use ffmpeg to convert the bytes to a format that we can use
"""

# # This part can be ignored when using a different protocol.


def load_audio(data: Union[str, bytes]) -> np.array:
    def read_file_to_bytes(data, conn):
        # Seen as serialized bytes.
        # This is the most important part; sending CBOR-seralized data to the ffmpeg_reader process
        conn.send(data)

    def ffmpeg_reader(conn, queue):
        output = conn.recv()
        # Unpickle the data
        output = cbor.loads(output)

        temp_file = os.path.join(tempfile.gettempdir(), "temp.wav")

        with wave.open(temp_file, "wb") as f:
            f.setparams(
                (
                    output["nchannels"],
                    output["sampwidth"],
                    output["framerate"],
                    output["nframes"],
                    output["comptype"],
                    output["compname"],
                )
            )
            f.writeframes(output["frames"])
        out, _ = (
            ffmpeg.input(temp_file)
            .output("-", format="s16le", acodec="pcm_s16le", ac=1, ar=16000)
            .run(cmd=["ffmpeg", "-nostdin"], capture_stdout=True, capture_stderr=True)
        )
        out = np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0

        queue.put(out)

    conn1, conn2 = multiprocessing.Pipe(True)
    queue: multiprocessing.Queue = multiprocessing.Queue()

    # Transform data into bytes if it is a string
    if isinstance(data, str):
        with wave.open(data, "rb") as f:
            out = dict(
                nframes=f.getnframes(),
                sampwidth=f.getsampwidth(),
                framerate=f.getframerate(),
                nchannels=f.getnchannels(),
                comptype=f.getcomptype(),
                compname=f.getcompname(),
                frames=f.readframes(f.getnframes()),
            )
            data = cbor.dumps(out)

    read = multiprocessing.Process(target=read_file_to_bytes, args=(data, conn1))
    reader = multiprocessing.Process(target=ffmpeg_reader, args=(conn2, queue))
    read.start()
    reader.start()

    return queue.get()



# Copyright 2022 Mithril Security. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import enum
import hashlib
import importlib
import os
from typing import Optional
from typing_extensions import Self
from dataclasses import dataclass
import sgx_dcap_quote_verify
from sgx_dcap_quote_verify import VerificationStatus
import toml
from pathlib import Path


@dataclass
class Collateral:
    version: int
    pck_certificate: str
    pck_crl_issuer_chain: str
    pck_signing_chain: str
    root_ca_crl: str
    pck_crl: str
    tcb_info: str
    tcb_info_issuer_chain: str
    qe_identity: str
    qe_identity_issuer_chain: str

    def __post_init__(self):
        for name, field_type in self.__annotations__.items():
            if not isinstance(self.__dict__[name], field_type):
                current_type = type(self.__dict__[name])
                raise TypeError(
                    f"The field `{name}` was assigned by {current_type} instead of {field_type}"
                )


class AttestationError(Exception):
    """This exception is raised when the attestation is invalid (enclave
    settings mismatching, debug mode unallowed...).

    Used as master exception for all other sub exceptions on the attestation
    validation
    """

    pass


class QuoteValidationError(Exception):
    """This exception is raised when the returned quote is invalid (TCB
    outdated, not signed by the hardware provider...).

    Used as master exception for all other sub exceptions on the quote
    validation
    """

    pass


class EnclaveHeldDataError(QuoteValidationError):
    """This exception is raised when the enclave held data expected does not
    match the one in the quote.

    The expected enclave held data in BlindAI is an SHA-256 hash of the enclave certificate to avoid man-in-the-middle attacks
    Args:
        expected_hash (str): Enclave held data hash expected
        got_hash (str): Enclave held data hash obtained from the quote's report
    """

    def __init__(self, expected: bytes, got: bytes):
        self.expected_hash = expected
        self.measured_hash = got
        super().__init__(
            f"Hash of enclave held data doesn't match with the report data from the quote. Expected {expected.hex()}, got {got.hex()} instead."
        )

    pass


class IdentityError(QuoteValidationError):
    """This exception is raised when the enclave code digest (MRENCLAVE is SGX terminology) does not match the digest provided in the manifest
    Args:
        expected_hash (str): hash from manifest
        got_hash (str): hash obtained from the quote's report
    """

    def __init__(self, expected: bytes, got: bytes):
        self.expected_hash = expected
        self.got_hash = got
        super().__init__(
            f"Error invalid MRENCLAVE. Expected {expected.hex()}, got {got.hex()} instead."
        )


def validate_attestation(
    quote: bytes,
    collateral: Collateral,
    enclave_held_data: bytes,
    manifest_path: Optional[Path] = None,
    use_cloud_manifest: bool = False,
):
    """Verifies if the enclave evidence is valid.

    * Validates if the quote is trustworthy (issued by an approved Intel CPU) with the
        attestation collateral using SGX Quote Verification Library.
    * Validates if the SHA256 hash of Enclave Held Data (EHD) matches the first 32 bytes
        of reportData field in the enclave quote. After this check
        we can be sure that the EHD bytes are endorsed by the enclave.
    Args:
        quote (bytes): SGX quote
        attestation_collateral (SgxCollateral): SGX collateral needed to assess the validity of the quote
            (collateral is signed by Intel)
        enclave_held_data (bytes): Enclave held data
    Raises:
        QuoteValidationError: The quote could not be validated.
        EnclaveHeldDataError: The enclave held data expected does not match the one in the quote. The expected enclave held data in BlindAI is a certificate to avoid man-in-the-middle attacks.
        NotAnEnclaveError: The enclave claims are not validated by the hardware provider, meaning that the claims cannot be verified using the hardware root of trust.
    Returns:
        -
    """

    # TODO: Handle the case where the retuned quote status is STATUS_TCB_SW_HARDENING_NEEDED
    # We must do more cautious checks in this case in order to determine whether or not to accept the quote

    trusted_root_ca_certificate = importlib.resources.read_text(  # type: ignore
        __package__, "Intel_SGX_Provisioning_Certification_RootCA.pem"
    )

    attestation_result = sgx_dcap_quote_verify.verify(
        trusted_root_ca_certificate=trusted_root_ca_certificate,
        pck_certificate=collateral.pck_certificate,
        pck_signing_chain=collateral.pck_signing_chain,
        root_ca_crl=collateral.root_ca_crl,
        intermediate_ca_crl=collateral.pck_crl,
        tcb_info=collateral.tcb_info,
        tcb_signing_chain=collateral.tcb_info_issuer_chain,
        quote=quote,
        qe_identity=collateral.qe_identity,
        expiration_date=datetime.now(),
    )
    if attestation_result.pck_certificate_status != VerificationStatus.STATUS_OK:
        raise QuoteValidationError(
            f"Invalid PCK certificate status {attestation_result.pck_certificate_status.name}"
        )

    if attestation_result.tcb_info_status != VerificationStatus.STATUS_OK:
        raise QuoteValidationError(
            f"Invalid TCB info Status {attestation_result.tcb_info_status.name}"
        )

    if attestation_result.qe_identity_status != VerificationStatus.STATUS_OK:
        raise QuoteValidationError(
            f"Invalid QE identity status {attestation_result.qe_identity_status.name}"
        )

    if attestation_result.quote_status not in [
        VerificationStatus.STATUS_OK,
        VerificationStatus.STATUS_TCB_SW_HARDENING_NEEDED,
    ]:
        raise QuoteValidationError(
            f"Invalid quote status {attestation_result.quote_status.name}",
        )

    assert attestation_result.enclave_report is not None
    if (
        hashlib.sha256(enclave_held_data).digest()
        != attestation_result.enclave_report.report_data[:32]
    ):
        raise EnclaveHeldDataError(
            expected=hashlib.sha256(enclave_held_data).digest(),
            got=attestation_result.enclave_report.report_data[:32],
        )

    if manifest_path is not None and use_cloud_manifest is True:
        raise TypeError(
            "Input inconsistency. You cannot set a manifest path and ask to use the cloud manifest in the same time"
        )

    if manifest_path is None:
        if use_cloud_manifest is False:
            manifest = EnclaveManifest.from_str(
                importlib.resources.read_text(__package__, "manifest.toml")  # type: ignore
            )
        else:
            manifest = EnclaveManifest.from_str(
                importlib.resources.read_text(__package__, "manifest_cloud.toml")  # type: ignore
            )
    else:
        if not isinstance(manifest_path, Path):
            raise ValueError("manifest_path should be a pathlib.Path")
        manifest = EnclaveManifest.from_str(manifest_path.read_text())

    if attestation_result.enclave_report.mr_enclave != manifest.mr_enclave:
        raise IdentityError(
            expected=manifest.mr_enclave,
            got=attestation_result.enclave_report.mr_enclave,
        )

    # Attributes data structures is made of two fields : two 64-bits bitvectors
    enclave_attributes_flags = int.from_bytes(
        attestation_result.enclave_report.attributes[0:8], byteorder="little"
    )
    enclave_attributes_xfrm = int.from_bytes(
        attestation_result.enclave_report.attributes[8:16], byteorder="little"
    )

    if (
        SgxAttributesFlags.DEBUG in SgxAttributesFlags(enclave_attributes_flags)
        and not manifest.allow_debug
    ):
        raise AttestationError(
            "Enclave is running in debug mode but the manifest forbids debug mode",
        )

    if (
        enclave_attributes_flags & manifest.attributes_mask_flags
        != manifest.attributes_flags | SgxAttributesFlags.INIT
    ):
        raise AttestationError(
            "SGX_ATTRIBUTES.FLAGS does not conform to the enclave manifest",
            SgxAttributesFlags(enclave_attributes_flags),
            SgxAttributesFlags(manifest.attributes_flags),
        )

    if (
        enclave_attributes_xfrm & manifest.attributes_mask_xfrm
        != manifest.attributes_xfrm
    ):
        raise AttestationError(
            "SGX_ATTRIBUTES.XFRM does not conform to the enclave manifest",
            enclave_attributes_xfrm,
            manifest.attributes_xfrm,
        )
    enclave_misc_select = attestation_result.enclave_report.misc_select.value
    if enclave_misc_select & manifest.misc_mask != manifest.misc_select:
        raise AttestationError(
            "SGX_MISC_SELECT does not conform to the enclave manifest",
            SgxMiscSelect(attestation_result.enclave_report.misc_select.value),
            SgxMiscSelect(manifest.misc_select),
        )


def hex_to_u64(hex_string: str) -> int:
    int_value = int(hex_string, 16)
    if int_value < 0 or int_value >= 2**64:
        raise ValueError("Not a valid unsigned 64-bit integer")
    return int_value


def hex_to_u32(hex_string: str) -> int:
    int_value = int(hex_string, 16)
    if int_value < 0 or int_value >= 2**32:
        raise ValueError("Not a valid unsigned 32-bit integer")
    return int_value


@dataclass
class EnclaveManifest:
    mr_enclave: bytes
    allow_debug: bool

    attributes_flags: int  # 64-bits bitvector
    attributes_mask_flags: int  # 64-bits bitvector

    attributes_xfrm: int  # 64-bits bitvector
    attributes_mask_xfrm: int  # 64-bits bitvector

    misc_select: int  # 32-bits bitvector
    misc_mask: int  # 32-bits bitvector

    def __post_init__(self):
        for name, field_type in self.__annotations__.items():
            if not isinstance(self.__dict__[name], field_type):
                current_type = type(self.__dict__[name])
                raise TypeError(
                    f"The field `{name}` was assigned by `{current_type}` instead of `{field_type}`"
                )

    @staticmethod
    def from_str(s: str) -> "EnclaveManifest":
        """Load a manifest from the content of a manifest.toml
        Args:
            s (str): The content of the manifest.
        Returns:
            manifest: The manifest.
        """
        return EnclaveManifest.from_dict(toml.loads(s))

    @staticmethod
    def from_file(path: str) -> "EnclaveManifest":
        """Load a manifest from a file.

        Args:
            path (str): The path of the file.
        Returns:
            manifest: The manifest.
        """
        return EnclaveManifest.from_dict(toml.load(path))

    @staticmethod
    def from_dict(obj: dict) -> "EnclaveManifest":
        """Load an enclave manifest from the dict obtained after manifest.toml
        decoding.

        Args:
            obj (dict): The dict.
        Returns:
            EnclaveManifest:
        """
        mr_enclave = bytes.fromhex(obj["mr_enclave"])
        if len(mr_enclave) != hashlib.sha256().digest_size:
            raise ValueError("Invalid mr_enclave in Manifest")
        return EnclaveManifest(
            mr_enclave=mr_enclave,
            allow_debug=obj["allow_debug"],
            attributes_flags=hex_to_u64(obj["attributes_flags_hex"]),
            attributes_mask_flags=hex_to_u64(obj["attributes_mask_flags_hex"]),
            attributes_xfrm=hex_to_u64(obj["attributes_xfrm_hex"]),
            attributes_mask_xfrm=hex_to_u64(obj["attributes_mask_xfrm_hex"]),
            misc_mask=hex_to_u32(obj["misc_mask_hex"]),
            misc_select=hex_to_u32(obj["misc_select_hex"]),
        )


class SgxAttributesFlags(enum.IntFlag):
    """ATTRIBUTES_FLAGS are part the SGX REPORT ATTRIBUTES_FLAGS is a 64 bits
    long bit field.

    For more information about the structure, refer to Intel Reference
    available at
    <https://www.intel.com/content/dam/www/public/us/en/documents/manuals/64-ia-32-architectures-software-developer-vol-3d-part-4-manual.pdf>
    """

    # This bit indicates if the enclave has been initialized by EINIT. It must be cleared when loaded as
    # part of ECREATE. For EREPORT instruction, TARGET_INFO.ATTRIBUTES[ENIT] must always be 1 to
    # match the state after EINIT has initialized the enclave
    INIT = 1 << 0
    # If set the enclave permit debugger to read and write enclave data using EDBGRD and EDBGWR.
    DEBUG = 1 << 1
    # If set, enclave runs in 64-bit mode.
    MODE64BIT = 1 << 2
    # Must be Zero.
    RESERVED = 1 << 3
    # Provisioning Key is available from EGETKEY.
    PROVISIONKEY = 1 << 4
    # EINIT token key is available from EGETKEY.
    EINITTOKENKEY = 1 << 5
    # The remaining bits 63:6 are reserved for future use


class SgxMiscSelect(enum.IntFlag):
    """The bit vector of MISCSELECT selects which extended information is to be
    saved in the MISC region of the SSA frame when an AEX is generated.

    For more information about the structuren, refer to Intel Reference
    available at
    <https://www.intel.com/content/dam/www/public/us/en/documents/manuals/64-ia-32-architectures-software-developer-vol-3d-part-4-manual.pdf>
    """

    # Report information about page fault and general protection exception that occurred inside an enclave.
    EXINFO = 1 << 0
    # The remaining bits 31:1 are reserved for future use and must be 0 in the meantime.



__all__ = [
    "connect",
    "RunModelResponse",
    "UploadResponse",
    "Tensor",
    "TensorInfo",
    "ModelDatumType",
    "AttestationError",
    "QuoteValidationError",
    "EnclaveHeldDataError",
    "IdentityError",
    "testing",
]

from .client import (
    connect,
    RunModelResponse,
    UploadResponse,
    Tensor,
    TensorInfo,
    ModelDatumType,
)

from ._dcap_attestation import (
    AttestationError,
    QuoteValidationError,
    EnclaveHeldDataError,
    IdentityError,
)
from . import testing



# Copyright 2022 Mithril Security. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import cryptography.x509
from cryptography.hazmat.primitives import serialization
import torch
import os


def strip_https(url: str) -> str:
    return re.sub(r"^https:\/\/", "", url)


def cert_der_to_pem(cert: bytes) -> bytes:
    return cryptography.x509.load_der_x509_certificate(cert).public_bytes(
        serialization.Encoding.PEM
    )


def fetch_whisper_tiny_20_tokens():
    # TODO: Urgent
    # Remove this implementation and actually convert the model to ONNX

    # We convert the method to ONNX
    # For the mean time, we will return a downloaded whisper model

    # Add filename to current directory

    filename = "whisper-tiny-20-tokens.onnx"
    filename = os.path.join(os.path.dirname(__file__), filename)

    # Check if the file doesn't exist, then download it
    if not os.path.exists(filename):
        os.system(f"gdown 1wqg1F0UkEdm3KB7n1BjfRLHnzKU2-G5S --output {filename}")

    # Return the absolute path of the file
    return os.path.abspath(filename)



import whisper
from typing import Optional, Union
from .utils import fetch_whisper_tiny_20_tokens
from .client import BlindAiConnection, connect
from transformers import WhisperProcessor
import torch
from ._preprocess_audio import load_audio

DEFAULT_BLINDAI_ADDR = "4.246.205.63"
DEFAULT_WHISPER_MODEL = "tiny.en"
DEFAULT_TEE_OPTIONS = ["sgx", "nitro"]
DEFAULT_TEE = "sgx"
DEFAULT_TRANSFORMER = f"openai/whisper-{DEFAULT_WHISPER_MODEL}"
DEFAULT_MODEL_HASH = "ff63656d9b09514efbb38b4b69324280a86b55df5e3a2268cb79e812d8c7b863"


def _preprocess_audio(file: Union[str, bytes]) -> torch.Tensor:
    """
    Preprocess audio file to be used with Whisper model.

    Args:
        file: str
            Audio file to preprocess

    Returns:
        torch.Tensor:
            The preprocessed audio file
    """
    # Load audio file
    audio = load_audio(file).flatten()

    # Pad or trim ndarray to [80, 3000]
    audio = whisper.pad_or_trim(audio)

    # Convert loaded audio to log_mel_spectrogram
    return whisper.log_mel_spectrogram(audio).unsqueeze(0)


def _get_connection(
    connection: Optional["BlindAiConnection"], tee: Optional[str]
) -> "BlindAiConnection":
    """
    Get the BlindAI connection object.

    Args:
        connection: Optional[BlindAiConnection]
            The BlindAI connection object

    Returns:
        BlindAiConnection:
            The BlindAI connection object
    """
    if connection is None:
        connection = connect(
            DEFAULT_BLINDAI_ADDR,
            hazmat_http_on_unattested_port=True,
            use_cloud_manifest=True,
        )

    return connection


class Audio:
    @classmethod
    def transcribe(
        cls,
        file: Union[str, bytes],
        model: str = DEFAULT_WHISPER_MODEL,
        connection: Optional["BlindAiConnection"] = None,
        tee: Optional[str] = DEFAULT_TEE,
    ) -> str:
        """
        BlindAI Whisper API which converts speech to text based on the model passed.

        Args:

            file: str, bytes
                Audio file to transcribe. It may also receive serialized bytes of wave file.
            model: str
                The Whisper model. Defaults to "medium".
            connection: Optional[BlindAiConnection]
                The BlindAI connection object. Defaults to None.
            tee: Optional[str]
                The Trusted Execution Environment to use. Defaults to "sgx". Unused, at the moment.
        Returns:
            Dict:
                The transcription object containing, text and the tokens
        """

        # Check which TEE to use
        if tee not in DEFAULT_TEE_OPTIONS:
            raise ValueError(f"tee must be one of {DEFAULT_TEE_OPTIONS}")

        # Preprocess audio file
        input_mel = _preprocess_audio(file)

        # Get BlindAI connection object
        with _get_connection(connection, tee) as conn:
            # Run ONNX model with `input_array` on BlindAI server
            res = conn.run_model(model_hash=DEFAULT_MODEL_HASH, input_tensors=input_mel)

            # Convert each output BlindAI Tensor object into PyTorch Tensor
            res = [t.as_torch() for t in res.output]  # type: ignore

            # Load the Whisper Transformer object.
            tokenizer = WhisperProcessor.from_pretrained(DEFAULT_TRANSFORMER)

            # Extract tokens from result
            tokens = res[0][0].numpy()  # type: ignore

            # Use transform to decode tokens
            text = tokenizer.batch_decode(tokens, skip_special_tokens=True).pop()

            return text
